/*
RANDOM SPRITE GENERATOR
Based on the Javascript at
http://img.uninhabitant.com/spritegen.html
*/

// tile control parameters
final int tiles = 16;
final int size = 16;
final int spacing = 2;
final int zoom = 2;

// tile mirror parameters
final float x_mirror_prob = 1.0;
final float y_mirror_prob = 0.0;

// pixel solidity parameters
final float x_scale_solid = 8.0 / size;
final float y_scale_solid = 8.0 / size;
final float min_prob = 0.2;
final float max_prob = 0.8;

// pixel color parameters
final float x_scale_color = 8.0 / size;
final float y_scale_color = 8.0 / size;
final int num_colors = 3;

// pixel post-processing parameters
final float despeckle_prob = size > 8 ? 1.0 : 0.5;
final float despur_prob = size > 8 ? 1.0 : 0.5;
final int min_foreground = size * size / 16;

// show seed numbers?
boolean debug_seed = false;

// on-screen palette
final int pal_height = 2 * zoom;

// color pattern
int color_pattern = 0;
int color_shift = 0;

// active palette
int active_palette = 1; 

// starting sprite seed for the page
int sprite_seed = 0;

// tile positioning
final int start = spacing / 2;
final int advance = size + spacing;

// sprite opacity mode
boolean background_opaque = false;

// render buffer
PGraphics graphics;

void setup()
{
  size(tiles * advance * zoom, tiles * advance * zoom + pal_height);
  noSmooth();
  noStroke();
  noLoop();
  textAlign(LEFT, TOP);
  textSize(9);
  
  graphics = createGraphics(tiles * advance, tiles * advance);
  
  load_palettes("data/palettes.json");

  println("");
  println("Controls:");
  println("up and down arrows change starting seed");
  println("left and right arrows change color pattern");
  println("; and ' shift colors");
  println("[ and ] change palette");
  println("S toggles sprite seed overlays");
  println("T writes an uncompressed .TGA image file");
  println("P writes a compressed .PNG image file");
  println("B switches between transparent and opaque background");
  println("number keys change opacity falloff");
  println("F1 shows sprite images");
  println("F2 shows noise pattern used for opacity");
  println("F3 shows probability falloff pattern");
  println("F4 shows opacity calculation");
  println("F5 shows noise pattern used for color");
  println("");
  println("Current state:");
  println(String.format("palette %d: %s", active_palette, palette_name[active_palette]));
  println(String.format("opacity falloff: %s", prob_name[falloff_mode]));
  println(String.format("seed start: %d", sprite_seed));
  println(String.format("color pattern: %d", color_pattern));
  println(String.format("%s background", background_opaque ? "opaque" : "transparent"));
  println("");
  println("Ready...");
}

void draw()
{
  // get the palette
  final color[] palette = palettes[active_palette];

  // set background
  background(palette[0]);

  // generate solid probability table
  // (shared between tiles)
  final float[][] prob_table = compute_falloff(falloff_mode, size);
  
  graphics.loadPixels();
  
  // clear the render buffer
  final color fill = background_opaque ? palette[0] : color(0, 0, 0, 0);
  for (int i = 0; i < graphics.pixels.length; ++i)
  {
    graphics.pixels[i] = fill;
  }
  
  // for each sprite tile...
  for (int ty = 0; ty < tiles; ++ty)
  {
    for (int tx = 0; tx < tiles; ++tx)
    {
      // draw a sprite into the render buffer
      final int x = start + tx * advance;
      final int y = start + ty * advance;
      final int seed = sprite_seed + ty * tiles + tx;
      generate_sprite(graphics, x, y, seed, prob_table, palette);
    }
  }
  graphics.updatePixels();
  
  // draw the render buffer with scaling
  image(graphics, 0, 0, graphics.width * zoom, graphics.height * zoom);

  // if drawing sprite seed overlays...
  if (debug_seed)
  {
    // print text in white    
    fill(255);
    
    for (int ty = 0; ty < tiles; ++ty)
    {
      for (int tx = 0; tx < tiles; ++tx)
      {
        // print sprite seed number
        final int x = start + tx * advance;
        final int y = start + ty * advance;
        final int seed = sprite_seed + ty * tiles + tx;
        text(seed, x * zoom, y * zoom);
      }
    }
  }

  if (pal_height > 0)
  {
    // draw the palette along the bottom
    final int pal_width = width / palette.length;
    for (int i = 0; i < palette.length; ++i)
    {
      fill(palette[i]);
      rect(i * pal_width, height - pal_height, pal_width, pal_height);
    }
  }
}

void saveImage(String filename)
{
  graphics.save(filename);
  println(String.format("saved \"%s\"", filename));
}

void keyPressed()
{
  switch(keyCode)
  {
  case java.awt.event.KeyEvent.VK_F1:
    debug_mode = DEBUG_NONE;
    println("draw sprites");
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_F2:
    debug_mode = DEBUG_NOISE;
    println("draw opacity noise values");
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_F3:
    debug_mode = DEBUG_PROB;
    println("draw opacity thresholds");
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_F4:
    debug_mode = DEBUG_SOLID;
    println("draw opacity calculation results");
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_F5:
    debug_mode = DEBUG_COLOR;
    println("draw color noise values");
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_0:
  case java.awt.event.KeyEvent.VK_1:
  case java.awt.event.KeyEvent.VK_2:
  case java.awt.event.KeyEvent.VK_3:
  case java.awt.event.KeyEvent.VK_4:
  case java.awt.event.KeyEvent.VK_5:
  case java.awt.event.KeyEvent.VK_6:
  case java.awt.event.KeyEvent.VK_7:
  case java.awt.event.KeyEvent.VK_8:
    falloff_mode = keyCode - java.awt.event.KeyEvent.VK_0;
    println(String.format("opacity falloff: %s", prob_name[falloff_mode]));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_S:
    debug_seed = !debug_seed;
    println(String.format("%s sprite seeds", debug_seed ? "show" : "hide"));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_P:
    saveImage(String.format("sprite_%d.png", sprite_seed));
    break;
  case java.awt.event.KeyEvent.VK_T:
    saveImage(String.format("sprite_%d.tga", sprite_seed));
    break;
  case java.awt.event.KeyEvent.VK_B:
    background_opaque = !background_opaque;
    println(String.format("%s background", background_opaque ? "opaque" : "transparent"));
    redraw();
    break;    
  case java.awt.event.KeyEvent.VK_UP:
    sprite_seed -= tiles * tiles;
    println(String.format("seed start: %d", sprite_seed));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_DOWN:
    sprite_seed += tiles * tiles;
    println(String.format("seed start: %d", sprite_seed));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_LEFT:
    --color_pattern;
    println(String.format("color pattern: %d", color_pattern));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_RIGHT:
    ++color_pattern;
    println(String.format("color pattern: %d", color_pattern));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_OPEN_BRACKET:
    active_palette = (active_palette + palettes.length - 1) % palettes.length;
    println(String.format("palette %d: %s", active_palette, palette_name[active_palette]));
    redraw();
    break; 
  case java.awt.event.KeyEvent.VK_CLOSE_BRACKET:
    active_palette = (active_palette + 1) % palettes.length;
    println(String.format("palette %d: %s", active_palette, palette_name[active_palette]));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_SEMICOLON:
    --color_shift;   
    println(String.format("color shift: %d", color_shift));
    redraw();
    break;
  case java.awt.event.KeyEvent.VK_QUOTE:
    ++color_shift;   
    println(String.format("color shift: %d", color_shift));
    redraw();
    break;
 }
}
