/*
SPRITE
generate a single sprite
*/

// sprite debug display mode
final int DEBUG_NONE = 0;
final int DEBUG_NOISE = 1;
final int DEBUG_PROB = 2;
final int DEBUG_SOLID = 3;
final int DEBUG_COLOR = 4;
int debug_mode = DEBUG_NONE;

// sprite color index indicating background
final byte sprite_background = -1;

// pick unique random colors from the palette
// the zeroth color is transparent (background)
void pick_random_colors(color[] colors, final color[] palette, final int seed)
{
  // reseed the generator so colors don't depend on sprite generation     
  randomSeed(seed);
  random(1);
  
  // pick random color entries from the palette
  for (int i = 0; i < colors.length; ++i)
  {
    final int entry = int(random(1, palette.length));
    colors[i] = palette[entry];
    for (int j = 0; j < i; ++j)
    {
      if (colors[i] == colors[j])
      {
        // got a duplicate; try again
        --i;
        break;
      }
    }
  }
}

// map sprite color indices to colors
void fill_pixels_sprite(PImage sprite, final int x_pos, final int y_pos, final byte[][] indices, final color[] colors)
{
  for (int iy = 0; iy < indices.length; ++iy)
  {
    for (int ix = 0; ix < indices[iy].length; ++ix)
    {
      // get color from sprite local palette
      byte index = indices[iy][ix];
      if (index != sprite_background)
      {
        // foreground pixel
        sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)] = colors[index];
      }
    }
  }
}

// map noise array values to intensities
void fill_pixels_noise(PImage sprite, final int x_pos, final int y_pos, final int x_size, final int y_size, final float[][] noise_table, final float x_scale, final float y_scale)
{
  for (int iy = 0; iy < y_size; ++iy)
  {
    for (int ix = 0; ix < x_size; ++ix)
    {
      // evaluate noise function at this pixel
      final float value = eval_noise(noise_table, ix * x_scale, iy * y_scale);
      
      // show intensity
      sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)] = color(constrain(256 * value, 0, 255));
    }
  }
}

// map probability array values to intensities
void fill_pixels_probability(PImage sprite, final int x_pos, final int y_pos, final int x_size, final int y_size, final float[][] prob_table)
{
  for (int iy = 0; iy < y_size; ++iy)
  {
    for (int ix = 0; ix < x_size; ++ix)
    {
      // evaluate noise function at this pixel
      final float value = prob_table[iy][ix];
      
      // show intensity
      sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)] = color(constrain(256 * value, 0, 255));
    }
  }
}

// map solid computation values to signed intensities
void fill_pixels_solid(PImage sprite, final int x_pos, final int y_pos, final int x_size, final int y_size, final float[][] prob_table, final float[][] noise_table_solid, final float x_scale, final float y_scale)
{
  for (int iy = 0; iy < y_size; ++iy)
  {
    for (int ix = 0; ix < x_size; ++ix)
    {
      // get probability value for this pixel being solid        
      final float prob_solid = prob_table[iy][ix];

      // evaluate noise function at this pixel
      final float noise_solid = eval_noise(noise_table_solid, ix * x_scale_solid, iy * y_scale_solid);
      
      // show signed intensity
      final float value = prob_solid - noise_solid;
      if (value >= 0)
      {
        // foreground: blue -> cyan
        sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)] = color(0, constrain(256 * value, 0, 255), 255);
      }
      else
      {
        // background: black -> red
        sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)] = color(constrain(-256 * value, 0, 255), 0, 0);
      }
    }
  }
}

// generate a sprite
void generate_sprite(PImage sprite, final int x_pos, final int y_pos, final int seed, final float[][] prob_table, final color[] palette)
{
  // seed the generator
  randomSeed(seed);      
  random(1);

  // mirror?
  final boolean x_mirror = random(1) < x_mirror_prob;
  final boolean y_mirror = random(1) < y_mirror_prob;
 
  // pixel counts based on mirroring
  final int x_size = x_mirror ? size / 2 : size;
  final int y_size = y_mirror ? size / 2 : size;

  // create solid noise table
  final int x_size_solid = int(ceil(x_size * x_scale_solid) + 1);
  final int y_size_solid = int(ceil(y_size * y_scale_solid) + 1);
  float[][] noise_table_solid = new float[y_size_solid][x_size_solid];

  // create color noise table
  final int x_size_color = int(ceil(x_size * x_scale_color) + 1);
  final int y_size_color = int(ceil(y_size * y_scale_color) + 1);
  float[][] noise_table_color = new float[y_size_color][x_size_color];

  // fill color noise table
  // (this does not need to change when "rerolling" the sprite)
  generate_noise(noise_table_color, seed + color_pattern);

  // local color indices
  byte[][] indices = new byte[y_size][x_size];

  for (int attempt = 0; attempt < 16; ++attempt)
  {
    // fill solid noise table    
    generate_noise(noise_table_solid, seed + attempt * 65536);
  
    // for each pixel...
    for (int iy = 0; iy < y_size; ++iy)
    {
      for (int ix = 0; ix < x_size; ++ix)
      {
        // get probability value for this pixel being solid        
        final float solid_prob = prob_table[iy][ix];

        // evaluate solid noise value
        final float noise_solid = eval_noise(noise_table_solid, ix * x_scale_solid, iy * y_scale_solid);
        
        // if the pixel is solid...
        if (noise_solid < solid_prob)
        {
          // evaluate color noise value
          float noise_color = eval_noise(noise_table_color, ix * x_scale_color, iy * y_scale_color);
          
          // if using the same noise data for solid and color (color pattern zero, equal scales)...
          if (noise_color == noise_solid)
          {
            // divide by the solid probability to ensure it covers the full range of colors
            noise_color /= solid_prob;
          }
          
          // set foreground color index based on the color noise value
          indices[iy][ix] = byte(constrain(num_colors * noise_color, 0, num_colors - 1));
        }
        else
        {
          // set background color index
          indices[iy][ix] = sprite_background;
        }
      }
    }

    // post-process to eliminate visual noise
    for (int iy = 0; iy < y_size; ++iy)
    {
      for (int ix = 0; ix < x_size; ++ix)
      {
        // skip background pixels
        if (indices[iy][ix] == sprite_background)
          continue;
          
        // count foreground neighbors
        int neighbors = 0;
        for (int ny = max(iy - 1, 0); ny < min(iy + 2, y_size); ++ny)
        { 
          for (int nx = max(ix - 1, 0); nx < min(ix + 2, x_size); ++nx)
          {
            // skip self unless on a mirror border
            if (nx == ix && ny == iy && !(x_mirror && nx == x_size-1) && !(y_mirror && ny == y_size-1))
              continue;
              
            // add neighbor if foreground
            if (indices[ny][nx] != sprite_background)
              ++neighbors;
          }
        }
        switch (neighbors)
        {
        case 0: // despeckle
          if (random(1) < despeckle_prob)
            indices[iy][ix] = sprite_background;
          break;
        case 1: // despur
          if (random(1) < despur_prob)
            indices[iy][ix] = sprite_background;
          break;
        }
      }
    }
    
    // count foreground pixels
    // (taking mirroring into account)
    int foreground = 0;
    for (int iy = 0; iy < y_size; ++iy)
    {
      for (int ix = 0; ix < x_size; ++ix)
      {
        if (indices[iy][ix] != sprite_background)
          ++foreground;
      }
    }
    if (x_mirror)
    {
      foreground *= 2;
    }
    if (y_mirror)
    {
      foreground *= 2;
    }
    
    // if the sprite has enough pixels...
    if (foreground >= min_foreground)
    {
      // use this image
      break;
    }
    
    // try again with a new offset
    print(seed, "retry", attempt, "\n");
  }

  // fill sprite pixels
  switch (debug_mode)
  {
  case DEBUG_NONE:
    // sprite local palette
    color[] colors = new color[num_colors];
    pick_random_colors(colors, palette, seed + color_shift);
    fill_pixels_sprite(sprite, x_pos, y_pos, indices, colors);
    break;
    
  case DEBUG_NOISE:
    fill_pixels_noise(sprite, x_pos, y_pos, x_size, y_size, noise_table_solid, x_scale_solid, y_scale_solid);
    break;

  case DEBUG_PROB:
    fill_pixels_probability(sprite, x_pos, y_pos, x_size, y_size, prob_table);
    break;
    
  case DEBUG_COLOR:
    fill_pixels_noise(sprite, x_pos, y_pos, x_size, y_size, noise_table_color, x_scale_color, y_scale_color);
    break;
    
  case DEBUG_SOLID:
    fill_pixels_solid(sprite, x_pos, y_pos, x_size, y_size, prob_table, noise_table_solid, x_scale_solid, y_scale_solid); 
    break;
  }

  if (x_mirror)
  {
    // mirror x
    for (int iy = 0; iy < y_size; ++iy)
    {
      for (int ix = 0; ix < x_size; ++ix)
      {
        sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + size - 1 - ix)] = sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)];
      }
    }
  }
  
  if (y_mirror)
  {
    // mirror y
    for (int iy = 0; iy < y_size; ++iy)
    {
      for (int ix = 0; ix < size; ++ix)
      {
        sprite.pixels[(y_pos + size - 1 - iy) * sprite.width + (x_pos + ix)] = sprite.pixels[(y_pos + iy) * sprite.width + (x_pos + ix)];
      }
    }
  }
}
